import { Events } from 'yuzu-utils';
import { fn, IListener, IRef, setRefProps, eventHandlerFn, IStateLogger, IComponentConstructable } from '../types';
export declare type objDiffType = (match: Record<string, any>, obj: Record<string, any>, msg: (k: string, keys: string) => void) => void;
/**
 * `Component` is an extensible class constructor which provides the building block of Yuzu component system.
 *
 * It inherits methods from [**Events**](/packages/utils/api/events).
 *
 * ```js
 * const instance = new Component({ ... })
 * ```
 *
 * > **Lifecycle**
 * >
 * > | stage    | hooks     |
 * > |----------| --------- |
 * > | `create` | `created` |
 *
 * @class
 * @param {object} [options={}] Instance options
 * @property {boolean} $active `true` if the instance is mounted and initialized
 * @property {object} options Instance options (see [defaultOptions](#defaultOptions))
 * @property {object} state Instance state (see [setState](#setState))
 * @property {Element} $el The instance root DOM element (see [mount](#mount))
 * @property {Object.<string, Element>} $els  Object mapping references to component's child DOM elements (see `selectors` below)
 * @property {{ string: Component }} $refs Object mapping references to child components (see [setRef](#setRef))
 * @property {object} selectors Object mapping a child element's reference name and a CSS selector or custom function
 * @property {Object.<string, function|string>} listeners Object mapping DOM listeners and handlers (see [setListener](#setListener))
 * @property {Object.<string, function|string>} actions Object mapping state keys and functions to executed on state update
 * @returns {Component}
 */
export declare class Component<S = {}, O = {}> extends Events {
    static root?: string;
    static defaultOptions?: (self?: any) => Record<string, any>;
    /**
     * ```js
     * Component.YUZU_DATA_ATTR
     * ```
     * Component root element attribute marker.
     *
     * @static
     * @returns {object}
     */
    static YUZU_DATA_ATTR: string;
    /**
     * Marks yuzu components
     *
     * @static
     */
    static YUZU_COMPONENT: boolean;
    /**
     * ```js
     * Component.isComponent(obj)
     * ```
     *
     * Checks whether the passed-in value is a Component constructor.
     *
     * @static
     * @param {*} value
     * @returns {boolean}
     */
    static isComponent<T>(value: any): value is IComponentConstructable<T>;
    static displayName?: string;
    options: Readonly<O>;
    detached?: boolean;
    $active: boolean;
    $el: Element;
    $uid: string;
    $els: {
        [key: string]: Element | Element[];
    };
    $refs: {
        [key: string]: Component;
    };
    state: Readonly<S>;
    $context?: Record<string, any>;
    $parent?: Component;
    selectors?: Record<string, string | ((el: Element, options: O) => Element | Element[])>;
    listeners?: Record<string, string | eventHandlerFn>;
    actions?: Record<string, string | fn>;
    $refsStore: Map<string, Component>;
    $listeners: Map<eventHandlerFn, IListener>;
    readyState?: (current: Readonly<S>, prev: Readonly<S>) => boolean;
    $warn: (msg: string, ...args: any[]) => void;
    $$logStart?: fn;
    $$logEnd?: fn;
    $$logger?: IStateLogger<Component, Readonly<S>>;
    /**
     * ```js
     * this.readyState(state, prevState)
     * ```
     *
     * Optional method used to delay the execution of [`ready`](#ready) after the state satisfies to a given condition.
     *
     * Will be executed at every state change until it returns `true`.
     *
     * @name readyState
     * @public
     * @type {function}
     * @param {object} state The current state
     * @param {object} prevState The previous state
     * @returns {boolean}
     */
    /**
     * Component constructor
     */
    constructor(options?: Partial<O>);
    /**
     * ```js
     * this.defaultOptions()
     * ```
     * Returns an object with default options.
     *
     * @param {Component} self The component instance itself
     * @returns {object}
     */
    defaultOptions(self?: this): O;
    /**
     * ```js
     * mount(el, [state])
     * ```
     *
     * Mounts a component instance on a DOM element and initializes it.
     *
     * ?> To prevent initialization and just mount the component set `state` to `null`.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `mount` | `beforeMount` <sup>(1)</sup>, `mounted` |
     *
     * 1. Just after attaching the root element (`this.$el`) but before any listener and selector registration.
     *
     * @param {string|Element} el Component's root element
     * @param {object|null} [state={}] Initial state
     * @returns {Component}
     */
    mount(el: string | Element, state?: Partial<S> | null): this;
    /**
     * ```js
     * init([state])
     * ```
     * Initializes the component instance.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `init` | `initialize`, `ready` |
     *
     * @param {object|null} [state={}] Initial state
     * @returns {Component}
     */
    init(state?: Partial<S>): this;
    /**
     * Lifecycle hook called on instance creation.
     *
     * At this stage just the instance options (`this.options`) are initialized.
     *
     *  Overwrite this method with custom logic in your components.
     *
     * ?> Use this hook to tap as early as possible into the component's properties. For example to set a dynamic initial state.
     */
    created(): void;
    /**
     * Lifecycle hook called just before mounting the instance onto the root element.
     *
     * At this stage the component is already attached to its root DOM element.
     *
     * Overwrite this method with custom logic in your components.
     */
    beforeMount(): void;
    /**
     * Lifecycle hook called when the instance gets mounted onto a DOM element.
     *
     * At this stage both children elements (`this.$els.*`) and event listeners configured in `this.listeners` have been setup.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> Use this method when you need to work with the DOM or manage any side-effect that requires the component to be into the DOM.
     */
    mounted(): void;
    /**
     * Lifecycle hook called before instance initialization.
     *
     * At this stage the state and state listeners are not yet been initialized.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> Use this method to set child components by [setRef](#setRef) and run any preparatory work on the instance.
     */
    initialize(): void;
    /**
     * Lifecycle hook called after instance initialization.
     *
     * At this stage State and event binding are already in place.
     *
     * Overwrite this method with custom logic in your components.
     *
     * ?> `ready` lifecycle can be delayed (_async ready_) by implementing a [`readyState`](packages/yuzu/#async-ready-state) method.
     */
    ready(): void;
    /**
     * Lifecycle hook called just before closing child refs.
     *
     * This hook is called just before destroying the instance. Every property, listener and state feature is still active.
     *
     * Overwrite this method with custom logic in your components.
     *
     * !> This is an async method. Return a promise in order to suspend the destroy process.
     */
    beforeDestroy(): void;
    /**
     * Returns an array of elements matching a CSS selector in the context of the component's root element
     *
     * @param selector {string} CSS selector to match
     * @return {Element[]}
     */
    findNodes(selector: string): Element[];
    /**
     * Returns the first element matching a CSS selector in the context of the component's root element
     *
     * @param selector {string} CSS selector to match
     * @return {Element}
     */
    findNode(selector: string): Element | null;
    /**
     * ```js
     * getState(key)
     * ```
     *
     * Returns a property of the state or a default value if the property is not set.
     *
     * ?> In ES6 environments you can use a [destructuring assignment](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Destructuring_assignment) instead: ` const { name = 'John'} = this.state`
     *
     *
     * @param {string} key Property to return
     * @param {*} def Default value
     * @returns {*}
     * @example
     * const instance = new Component().mount('#app', { a: 1 });
     * // instance.getState('a') === 1
     *
     * // instance.getState('b', false) === false
     */
    getState(key: keyof S, def?: any): any;
    /**
     * ```js
     * shouldUpdateState(string, currentValue, newValue)
     * ```
     * Executes a strict inequality comparison (`!==`) on the passed-in values and returns the result.
     *
     * !> This method is executed on every [`setState`](#setstate) call.
     *
     * You can overwrite this method with your own validation logic.
     *
     * @param {string} key State property name
     * @param {*} currentValue Value stored in the current state
     * @param {*} newValue New value
     * @returns {boolean}
     */
    shouldUpdateState(_key: keyof S, currentValue: any, newValue: any): boolean;
    /**
     * ```js
     * setState(updater, [silent])
     * ```
     *
     * Updates the internal instance state by creating a shallow copy of the current state and updating the passed-in keys.
     *
     * If the computed new state is different from the old one it emits a `change:<property>` event for every changed property
     * as well as a global `change:*` event.
     *
     * If the new value argument is a function, it will be executed with the current state as argument.
     * The returned value will be used to update the state.
     *
     * ?> To prevent an instance from emitting `change:` events set the second argument to `true` (silent update).
     *
     * @param {object|function} updater Defines which part of the state must be updated
     * @param {boolean} [silent=false] Update the state without emitting change events
     * @example
     * instance.on('change:a', (next, prev) => console.log(next, prev));
     * instance.setState({ a: 1 }); //emits 'change:a' -> logs undefined,1
     *
     * instance.setState({ a: 1 }); //nothing happens
     * instance.setState({ a: 2 }, true); //nothing happens, again...
     *
     * // use the current state to calculate its next value
     * instance.setState(({ a }) => ({a + 1}));
     */
    setState<K extends keyof S>(updater: (Pick<S, K> | Partial<S>) | ((prevState: Readonly<S>) => Pick<S, K> | Partial<S>), silent?: boolean): void;
    /**
     * ```js
     * replaceState(newState, [silent])
     * ```
     *
     * Replaces the current state of the instance with a completely new state.
     *
     * !> Note that this methods is un-affected by [`shouldUpdateState`](#shouldupdatestate).
     *
     * @param {object} newState The new state object
     * @param {boolean} [silent=false] Replace the state without emitting change events
     * @example
     * instance.replaceState({ a: 1 });
     * // instance.state.a === 1
     * instance.replaceState({ b: 2 });
     * // instance.state.b === 2
     * // instance.state.a === undefined
     */
    replaceState(newState: Record<string, any>, silent?: boolean): void;
    /**
     * ```js
     * broadcast(event, [...params])
     * ```
     *
     * Emits a `broadcast:<eventname>` event on every child component listed in `$refs`.
     *
     * @param {string} event Event name
     * @param {*[]} [params] Additional arguments to pass to the handler
     * @example
     * const child = new Component('#child');
     * child.on('broadcast:log', (str) => console.log(str));
     *
     * instance.setRef({ id: 'child', component: child });
     * instance.broadcast('log', 'test') // child component logs 'test'
     */
    broadcast(event: string, ...params: any[]): void;
    /**
     * ```js
     * setListener(string, handler)
     * ```
     *
     * Sets a DOM event listener.
     *
     * The first argument must be a string composed by an event name (ie `click`) and a CSS selector (`.element`)
     * separated by a space.
     *
     * If the CSS selector starts with `@` the listener will be attached to the
     * corresponding reference child element (`this.$els.<element>`), if any.
     *
     * @param {string} def Event and target element definition. Format `eventName [target]`
     * @param {function} handler Event handler
     * @example
     *
     * // attach a click handler to a child element
     * instance.setListener('click .button', () => ...)
     *
     * // attach a click handler to this.$els.btn
     * instance.setListener('click @btn', () => ...)
     *
     * // attach a click handler to this.$el
     * instance.setListener('click', () => ...)
     */
    setListener(def: string, handler: eventHandlerFn): void;
    /**
     * ```js
     * removeListeners()
     * ```
     *
     * Removes all DOM event listeners attached with `.setListener`.
     *
     */
    removeListeners(): void;
    /**
     * ```js
     * setRef(config, [props])
     * ```
     *
     * Attaches a reference to a child component.
     *
     * If a reference `id` is already attached, the previous one is destroyed and replaced with the new one.
     *
     * ?> This is an async method returning a promise.
     *
     * @param {object} config A child component configuration object
     * @param {string} config.id Reference id. Will be used to set a reference to the child component onto `this.$refs`
     * @param {component} config.component Component constructor or component instance
     * @param {string|HTMLElement} [config.el] Child component root element. This property is ignored if `config.component` is a component instance or a detached component constructor
     * @param {Object} [config.on] Child component event listeners. Format `{ 'eventname': handler }`
     * @param {*} config.* Any other property listed here will be passed to the constructor as option
     * @param {object} [props] Child component initial state
     * @returns {Promise}
     * @example
     * const parent = new Component('#root');
     *
     * class ChildComponent extends Component {}
     *
     * // as constructor
     * parent.setRef({
     *   id: 'child',
     *   component: ChildComponent,
     *   el: '#child',
     *   // other options here...
     * });
     *
     * // as instance
     * parent.setRef({
     *   id: 'child',
     *   component: new ChildComponent({ ... }).mount('#child', null) // <-- prevent component init
     * });
     *
     * // sync parent - child state
     * // updates child `parentCount` state whenever `parent.state.count` changes
     * parent.setRef({
     *   id: 'child',
     *   component: ChildComponent,
     *   el: '#child',
     *   // other options here...
     * }, {
     *   parentCount: (parentState) => parentState.count
     * });
     */
    setRef<C extends Component<any, any>>(refCfg: IRef<IComponentConstructable<C> | C | ((el: this['$el'], state: Readonly<S>) => C)>, props?: setRefProps<Component, this>): Promise<C>;
    /**
     * Destroys and detaches a specific child component by its reference `id` (as set in `setRef`).
     *
     * @param {string} id Child component reference id
     * @param {boolean} [detach=false] Remove the child component root element from the DOM
     * @returns {Promise}
     */
    destroyRef(id: string, detach?: boolean): Promise<void>;
    /**
     * Calls `.destroy()` on every child references and detaches them from the parent component.
     *
     * !> This is an async method returning a promise
     *
     * @returns {Promise}
     */
    destroyRefs(): Promise<void | void[]>;
    /**
     * ```js
     * destroy()
     * ```
     *
     * Detaches DOM events, instance's events and destroys all references as well.
     *
     * > **Lifecycle**
     * >
     * > | stage    | hooks     |
     * > |----------| --------- |
     * > | `destroy` | `beforeDestroy` |
     *
     *
     * !> This is an async method returning a promise
     *
     * @returns {Promise}
     */
    destroy(): Promise<void>;
}
